extern crate libc;
extern crate mraa_sys;

use std::fmt;

const MRAA_RESULT_CHECK_SYSLOG: mraa_sys::mraa_result_t = 10000;

#[derive(Debug)]
pub struct Error {
    inner: mraa_sys::mraa_result_t,
    operation: &'static str,
}

impl Error {
    fn new_check_syslog(operation: &'static str) -> Self {
        Self {
            inner: MRAA_RESULT_CHECK_SYSLOG,
            operation,
        }
    }

    pub fn raw_code(&self) -> mraa_sys::mraa_result_t {
        self.inner
    }

    pub fn operation(&self) -> &'static str {
        self.operation
    }
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "MRAA error on {}: ", self.operation)?;
        match self.inner {
            mraa_sys::mraa_result_t_MRAA_SUCCESS => write!(f, "Success"),
            mraa_sys::mraa_result_t_MRAA_ERROR_FEATURE_NOT_IMPLEMENTED => {
                write!(f, "Feature not implemented")
            }
            mraa_sys::mraa_result_t_MRAA_ERROR_FEATURE_NOT_SUPPORTED => {
                write!(f, "Feature not supported by HW")
            }
            mraa_sys::mraa_result_t_MRAA_ERROR_INVALID_PARAMETER => write!(f, "Parameter invalid"),
            mraa_sys::mraa_result_t_MRAA_ERROR_INVALID_RESOURCE => write!(f, "Resource invalid"),
            MRAA_RESULT_CHECK_SYSLOG => write!(f, "Unknown error, check syslog for error message"),
            unk => write!(f, "Unknown MRAA error: {}", unk),
        }
    }
}

impl std::error::Error for Error {}

pub type Result<T> = std::result::Result<T, self::Error>;

fn handle_result(inner: mraa_sys::mraa_result_t, operation: &'static str) -> self::Result<()> {
    if inner == mraa_sys::mraa_result_t_MRAA_SUCCESS {
        Ok(())
    } else {
        Err(Error { inner, operation })
    }
}

pub fn init() -> self::Result<()> {
    handle_result(unsafe { mraa_sys::mraa_init() }, "init")
}

// TODO: this crashes sometimes for some reason
pub unsafe fn deinit() {
    // unsafe { mraa_sys::mraa_deinit() };
    mraa_sys::mraa_deinit();
}

pub enum LogLevel {
    Emergency = 0,
    Alert = 1,
    Critical = 2,
    Error = 3,
    Warning = 4,
    Notice = 5,
    Info = 6,
    Debug = 7,
}

pub fn set_log_level(log_level: LogLevel) -> self::Result<()> {
    handle_result(
        unsafe { mraa_sys::mraa_set_log_level(log_level as i32) },
        "set_log_level",
    )
}

#[derive(Copy, Clone, PartialEq, Eq, Debug)]
pub enum GpioDirection {
    Out,
    In,
    OutHigh,
    OutLow,
}

impl GpioDirection {
    fn into_raw(self) -> mraa_sys::mraa_gpio_dir_t {
        use GpioDirection::*;
        match self {
            Out => mraa_sys::mraa_gpio_dir_t_MRAA_GPIO_OUT,
            In => mraa_sys::mraa_gpio_dir_t_MRAA_GPIO_IN,
            OutHigh => mraa_sys::mraa_gpio_dir_t_MRAA_GPIO_OUT_HIGH,
            OutLow => mraa_sys::mraa_gpio_dir_t_MRAA_GPIO_OUT_LOW,
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Debug)]
pub enum GpioEdge {
    NoEdge,
    Both,
    Rising,
    Falling,
}

impl GpioEdge {
    fn into_raw(self) -> mraa_sys::mraa_gpio_edge_t {
        use GpioEdge::*;
        match self {
            NoEdge => mraa_sys::mraa_gpio_edge_t_MRAA_GPIO_EDGE_NONE,
            Both => mraa_sys::mraa_gpio_edge_t_MRAA_GPIO_EDGE_BOTH,
            Rising => mraa_sys::mraa_gpio_edge_t_MRAA_GPIO_EDGE_RISING,
            Falling => mraa_sys::mraa_gpio_edge_t_MRAA_GPIO_EDGE_FALLING,
        }
    }
}

pub struct Gpio {
    inner: mraa_sys::mraa_gpio_context,
    isr_callback: Option<Box<dyn FnMut()>>,
}

impl Gpio {
    pub fn new(pin: i32) -> Result<Self> {
        let inner = unsafe { mraa_sys::mraa_gpio_init(pin) };
        if inner.is_null() {
            Err(Error::new_check_syslog("Gpio::new"))
        } else {
            Ok(Self {
                inner,
                isr_callback: None,
            })
        }
    }

    pub fn set_dir(&mut self, dir: GpioDirection) -> Result<()> {
        handle_result(
            unsafe { mraa_sys::mraa_gpio_dir(self.inner, dir.into_raw()) },
            "Gpio::set_dir",
        )
    }

    pub fn read(&mut self) -> Result<bool> {
        match unsafe { mraa_sys::mraa_gpio_read(self.inner) } {
            0 => Ok(false),
            1 => Ok(true),
            _ => Err(Error::new_check_syslog("Gpio::read")),
        }
    }

    pub fn set_isr<F>(&mut self, edge: GpioEdge, callback: F) -> Result<()>
    where
        F: FnMut() + 'static + Send + Sync,
    {
        if self.isr_callback.is_some() {
            let _ = self.remove_isr();
        }
        self.isr_callback = Some(Box::new(callback));

        let data = self.isr_callback.as_mut().unwrap().as_mut() as *mut _;
        let result = unsafe {
            mraa_sys::mraa_gpio_isr(
                self.inner,
                edge.into_raw(),
                Some(call_closure::<F>),
                data as *mut _,
            )
        };
        handle_result(result, "Gpio::set_isr")
    }

    pub fn remove_isr(&mut self) -> Result<()> {
        let result = unsafe { mraa_sys::mraa_gpio_isr_exit(self.inner) };
        self.isr_callback = None;
        handle_result(result, "Gpio::remove_isr")
    }
}

unsafe extern "C" fn call_closure<F>(data: *mut libc::c_void)
where
    F: FnMut(),
{
    let callback_ptr = data as *mut F;
    let callback = &mut *callback_ptr;
    callback();
}

impl Drop for Gpio {
    fn drop(&mut self) {
        unsafe { mraa_sys::mraa_gpio_close(self.inner) };
    }
}

pub struct Aio {
    inner: mraa_sys::mraa_aio_context,
}

impl Aio {
    pub fn new(pin: u32) -> Result<Self> {
        let inner = unsafe { mraa_sys::mraa_aio_init(pin) };
        if inner.is_null() {
            Err(Error::new_check_syslog("Aio::new"))
        } else {
            Ok(Self { inner })
        }
    }

    pub fn read(&mut self) -> Result<i32> {
        match unsafe { mraa_sys::mraa_aio_read(self.inner) } {
            -1 => Err(Error::new_check_syslog("Aio::read")),
            value => Ok(value),
        }
    }

    pub fn read_float(&mut self) -> Result<f32> {
        let value = unsafe { mraa_sys::mraa_aio_read_float(self.inner) };
        if value < 0.0 {
            Err(Error::new_check_syslog("Aio::read_float"))
        } else {
            Ok(value)
        }
    }
}

impl Drop for Aio {
    fn drop(&mut self) {
        unsafe { mraa_sys::mraa_aio_close(self.inner) };
    }
}

#[cfg(test)]
mod tests {
    #[test]
    fn it_works() {
        assert_eq!(2 + 2, 4);
    }
}
